from panda3d.core import Vec2, Point2
from MovingObject import MovingObject, normalizeDegrees
import math
# Huijie Bao
DT = 0.02
origin = Vec2(0, 0)

# this function can be used to display all the asteroid's values to the screen
def dumpAsteroids(rocks):
	print("Asteroids:")
	for rock in rocks:
		print(rock)
	print("------------------")
		
class ShipAI:
	# Called to create a new player
	def __init__(self, fieldSz, dt, turnRate, bulletSpeed, bulletRepeat, bulletRad, astSpeeds, astDiams,
					totalFrames, points):
		self.fieldSz = fieldSz
		self.dt = dt
		self.turnRate = turnRate
		self.bulletSpeed = bulletSpeed
		self.bulletRepeat = bulletRepeat
		self.bulletRad = bulletRad
		self.astSpeeds = astSpeeds
		
		self.astRads = []
		for diam  in astDiams:
			self.astRads.append(diam/2)
			
		self.totalFrames = totalFrames
		self.points = points
	
	# Called when a new game of asteroids is started.
	def newGame(self):
		self.pts = 0
		self.destroyingObjects = []
		#self.destroyedObjects = []
	# Called to get the player's next command.
	#
	# The return value must be a 2-tuple
	# of the form (fire, dir). Where dir = {negative, 0, positive} to indicate turn CCW, don't
	# turn, or turn CW. fire is a boolean value to indicate whether to fire a
	# bullet. If this is True, a bullet will be fire, but only if the bullet repeat
	# rate is not exceeded. To execute the "shield", return None
	#
	# The inputs are:
	# asteroids -- array of MovingObjects, one for each asteroid on screen
	# heading -- the ship's heading (0 degrees is down +x), degrees.
	# framesToFire -- num non-firing frames needed before next bullet.
	
	# calculate the efficiency to shoot this asteroid by (score-1)/abs(heading-direction)
	def getScoreOverCost(self,asteroid,heading):
		direction = normalizeDegrees(math.degrees(math.atan2(asteroid.pos.getY(),asteroid.pos.getX())))
		cost = abs(heading - direction)
		if cost ==0:
			return -1
		else:
			if asteroid.rad == 80:
				return 1/cost
			elif asteroid.rad == 40:
				return 3/cost
			else:
				return 9/cost
	
	# check if this MovingObject is in destroying by id
	def checkIDinDestroying(self,thisIid):
		for x in self.destroyingObjects:
			if x[0] == thisIid:
				return True
		return False
			
	# trace the target 
	def traceTarget(self,heading,target):
		direction = normalizeDegrees(math.degrees(math.atan2(target.pos.getY(),target.pos.getX())))
		if direction>heading:
			if direction-heading<=180:
				return (False, +1)
			else: 
				return (False, -1)
		elif direction<heading:
			if heading-direction<=180:
				return (False, -1)
			else:
				return (False, +1)
		else:
			return (False, 0)	
	
	def getAICommand(self, asteroids, heading, framesToFire):
		# dumpAsteroids(asteroids)
		# Fill this in for homework assignment
		
		# check destroying list		
		ind = 0
		while ind <len(self.destroyingObjects):
			self.destroyingObjects[ind][1] -= DT
			if self.destroyingObjects[ind][1] <=0:
				# destroy the object which should be destroyed from list
				self.destroyingObjects.pop(ind)
				ind -= 1
			ind += 1
		# end of checking destroying list
		
		# initialize tempship and tempbullet object
		thisShip =  MovingObject(-2,Point2(0,0),heading,0,10)
		temp_bullet = MovingObject(-1,Point2(0,0), heading,self.bulletSpeed,self.bulletRad)
	
		# build incoming asteroids list
		comingList = []#[index,time]
		#print(comingList)
		for x in range(0,len(asteroids)):
			if asteroids[x].collideWithObject(thisShip) !=None and asteroids[x].collideWithObject(thisShip) != (0,):#this ast is coming
				thisNeededTime = asteroids[x].collideWithObject(thisShip)
				#add this ast to list
				if(len(comingList)==0):#first element
					comingList.append([x,thisNeededTime])
					#print(comingList)
				else:#coming list has elements
					if thisNeededTime>comingList[len(comingList)-1][1]:
						comingList.append([x,thisNeededTime])
					else:
						for i in range(0,len(comingList)):
							if thisNeededTime < comingList[i][1]:
								comingList.insert(i, [x,thisNeededTime])
				#end of adding ast			
		# end of build incoming list

		# find closet asteroid to temp bullet
		# -1 means you will not shoot anything
		c_ast_b_id = -1 # id
		c_ast_time = 10000 # needed time

		for x in range(0,len(asteroids)):
			if asteroids[x].collideWithObject(temp_bullet) != None and asteroids[x].collideWithObject(temp_bullet) != (0,):
				if(asteroids[x].collideWithObject(temp_bullet)<c_ast_time ):
					shootedPoint_bullet = temp_bullet.pos+temp_bullet.vel*asteroids[x].collideWithObject(temp_bullet)
					boundOfScreen = 350 - self.bulletRad
					# if this shoot will work
					if shootedPoint_bullet.getX()>-boundOfScreen and shootedPoint_bullet.getX()<boundOfScreen and shootedPoint_bullet.getY()>-boundOfScreen and shootedPoint_bullet.getY()<boundOfScreen:
						c_ast_b_id = asteroids[x].id
						c_ast_time = asteroids[x].collideWithObject(temp_bullet)				
		# end of find closet asteroid to this bullet

		# if there are some threats		
		if len(comingList)>0:# there is sth in coming list
			
			#check if need open shield
			if comingList[0][1] <DT:
				return None
			#end of check if need open shield
			
			#when it's time to shoot
			target = asteroids[comingList[0][0]]
			# if the rest of time that ast will come is smaller than the time plane needed to rotate 180 degrees\
			if target.collideWithObject(temp_bullet) != None and target.collideWithObject(temp_bullet) != (0,) \
			and target.collideWithObject(temp_bullet)<=DT*180/4:
				# if target.id or c_ast_b_id is not in the destroying list
				if not self.checkIDinDestroying(target.id) and not self.checkIDinDestroying(c_ast_b_id):
					neededtime = comingList[0][1]
					shootedPoint_bullet = temp_bullet.pos+temp_bullet.vel*neededtime
					boundOfScreen = 350 - self.bulletRad
					# if this shoot will work
					if shootedPoint_bullet.getX()>-boundOfScreen and shootedPoint_bullet.getX()<boundOfScreen and shootedPoint_bullet.getY()>-boundOfScreen and shootedPoint_bullet.getY()<boundOfScreen:
						if(framesToFire<=0):
							if(c_ast_b_id!=-1 and c_ast_b_id!=target.id):
								self.destroyingObjects.append([c_ast_b_id,c_ast_time]) 
								return (True,0)
							else:
								self.destroyingObjects.append([target.id,neededtime]) 
								return (True,0)
						else:
							return self.traceTarget(heading, target)
					# if can not shoot successfully
					else:
						return self.traceTarget(heading, target)
				# if they are on the destroying list 
				else:
					return self.traceTarget(heading, target)
			else:
				return self.traceTarget(heading, target)

		#if there is no threat to you		
		else:
			# find the most efficient target
			scoreOverCost = 0
			priorIndex = -1 # most efficient index in asteroids
			for x in range(0,len(asteroids)):
				tempscore = self.getScoreOverCost(asteroids[x],heading)
				if tempscore == -1:
					scoreOverCost = 100000
					priorIndex = x
					break
				else:
					if(tempscore>scoreOverCost):
						scoreOverCost = tempscore
						priorIndex = x
			# end of find the most efficient target

			# in case nothing left in game
			if len(asteroids)==0 :	
				return (False,0)			
			# end in case
			
			if priorIndex!= -1 :
				target = asteroids[priorIndex]
				if not self.checkIDinDestroying(target.id) and not self.checkIDinDestroying(c_ast_b_id) and target.collideWithObject(temp_bullet)!=None and target.collideWithObject(temp_bullet)!=(0,):
					#have not shoot this target successfully, but it's coming
					neededtime = target.collideWithObject(temp_bullet)
					shootedPoint_bullet = temp_bullet.pos+temp_bullet.vel*neededtime
					boundOfScreen = 350-self.bulletRad
					if shootedPoint_bullet.getX()>-boundOfScreen and shootedPoint_bullet.getX()<boundOfScreen and shootedPoint_bullet.getY()>-boundOfScreen and shootedPoint_bullet.getY()<boundOfScreen:
						if(framesToFire<=0):						
							if(c_ast_b_id!=-1 and c_ast_b_id!=target.id):
								self.destroyingObjects.append([c_ast_b_id,c_ast_time]) 
								return (True,0) 
							else:
								self.destroyingObjects.append([target.id,neededtime]) 
								return (True,0)
						else:
							return self.traceTarget(heading, target)
					else:#can not shoot successfully
						#what to do?
						return self.traceTarget(heading, target)
				else:
					return self.traceTarget(heading, target)
			else:
				return self.traceTarget(heading, target)
		return (False, -1)		

	# This AI will simply point the ship toward the first asteroid in the
	# list. Shots are never fired by this AI.
	def trackingAI(self, asteroids, heading, framesToFire):
		if len(asteroids) == 0: return (False, 0)
		
		return (False, -1)
	
	# This AI will never turn. It will simply fire a bullet if the bullet
	# will intersect with the first asteroid in the list. It is acceptable to shoot
	# and miss when the intersection falls off the window.
	def opportunisticAI(self, asteroids, heading, framesToFire):
		if len(asteroids) == 0: return (False, 0)
		
		return (True, 0)

	# This AI will shoot at the first asteroid in the list, when a shot will be successful.
	# If a shot is not possible, the ship will turn toward the first asteroid in the list.
	def aggressiveAI(self, asteroids, heading, framesToFire):
		if len(asteroids) == 0: return (False, 0)
		return (True, 1)
	
